/*
 * beam.c
 * - code to beam texts across the TCP connection following a
 * special protocol
 * source file for pow-wow
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "powwow.h"
#include "term.h"
#include "xeval.h"

void bcopy();

#ifdef NOBCOPY
#define bcopy(a,b,c) my_memmove ((b), (a), (c))
#endif

extern time_t time(); /* ??? necessary ??? */

static char **split_line();
static void finish_edit();

editsess *edit_sess;	/* head of session list */

char edit_start[BUFSIZE]; /* messages to send to host when starting */ 
char edit_end[BUFSIZE];   /* or leaving editing sessions */ 

/*
 * Process editing protocol message from buf with len remaining chars.
 * Return number of characters used in the message.
 */
int process_message(buf, len)
char *buf; int len;
{
  int msglen, i, l, used;
  char *text;
    
  msglen = atoi(buf + 1);
  for(i = 1; i < len && isdigit(buf[i]); i++);
  if(buf[i] == '\r') i++;
  if(buf[i++] != '\0' || i >= len) 
   {
     PRINT_INT ("#Warning: protocol error\n");
     return len;
   }
  text = (char*)malloc(msglen);
  l = len - i;
  if(msglen < l) 
   {
     l = msglen;
     used = msglen + i;
   }
  else
     used = len;
  bcopy(buf + i, text, l);
  i = l;
  while(i < msglen) 
   {
     /*
      * read() might block here, but it should't be long. I should also
      * process any telnet protocol commands, but I don't care right now.
      */
     while((l = read(sockfd, text + i, msglen - i)) < 0 && errno == EINTR);
     printf("\rgot %d bytes out of %d", i, msglen); fflush(stdout);
     if(l < 0)
     	syserr("read message from socket");
     i += l;
   }
  printf("\rread all %d bytes.%s\n", msglen, clreoln);
  switch(*buf) 
   {
    case 'E':
     message_edit(text, msglen, 0, 0);
     break;
    case 'V':
     message_edit(text, msglen, 1, 0);
     break;
    default:
     printf("#Warning: received a funny message (0x%x)\n", *buf);
     free(text);
     break;
   }
  return used;
}

/*
 * start an editing session: process the EDIT/VIEW message
 * if view == 1, text will be viewed, else edited
 */
extern void message_edit(text, msglen, view, builtin)
char *text;
int msglen;
char view, builtin;
{
  char tmpname[LINELEN], buf[LINELEN];
  int tmpfd, key;
  int i;
  int childpid;
  editsess *s;
  char *editor, *descr;
  char **args;
  int waitforeditor;
  
  if(view) 
   {
     key = -1;
     i = 0;
   } 
  else 
   {
     if(text[0] != 'M') 
      {
	printf("#Warning: protocol error (message_edit, no M)\n");
	free(text);
	return;
      }
     for(i = 1; i < msglen && isdigit(text[i]); i++);
     if(text[i++] != '\n' || i >= msglen) 
      {
	printf("#Warning: protocol error (message_edit, no \\n)\n");
	free(text);
	return;
      }
     key = atoi(text + 1);
   }
  descr = text + i;
  while(i < msglen && text[i] != '\n') i++;
  if(i >= msglen) 
   {
     printf("#Warning: protocol error (message_edit, no desc)\n");
     free(text);
     return;
   }
  text[i++] = '\0';
  sprintf(tmpname, "/tmp/poww%d%d%d", getpid(), key, rand() >> 8);
  if((tmpfd = open(tmpname, O_WRONLY | O_CREAT, 0777)) < 0)
     syserr("create temporary edit file");
  if(write(tmpfd, text + i, msglen - i) < msglen - i)
     syserr("write to temp edit file");
  close(tmpfd);
  
  /* send a edit_start message (if wanted) */ 
  if ((!edit_sess) && (*edit_start)) {
     error = 0;
     parse_instruction(edit_start, 0, 0);
     history_done = 0;
  }

  s = (editsess*)malloc(sizeof(editsess));
  s->ctime = time((time_t*)NULL);
  s->oldsize = msglen - i;
  s->key = key;
  s->builtin = builtin;
  s->descr = strdup(descr);
  s->file = strdup(tmpname);
  
  free(text);
  if(view) 
   {
     if(!(editor = getenv("POWWOWPAGER")) && !(editor = getenv("PAGER")))
     	editor = "more";
   } 
  else 
   {
     if(!(editor = getenv("POWWOWEDITOR")) && !(editor = getenv("EDITOR")))
     	editor = "emacs";
   }
  if(editor[0] == '&') 
   {
     waitforeditor = 0;
     editor++;
   } 
  else 
   {
     waitforeditor = 1;
   }
  
  sprintf(buf, "%s %s", editor, tmpname);
  args = split_line(buf);
    
  if(waitforeditor) 
   {
     reset_terminal();
     /* ignore SIGINT since interrupting the child would interrupt us too,
      if we are in the same tty group */
     signal(SIGINT, SIG_IGN);
     signal(SIGCHLD, SIG_DFL);
   }
  childpid = fork();		/* let's get schizophrenic */
  switch(childpid) 
   {
    case 0:
     execvp(args[0], args);
     syserr("execvp");
     break;
    case -1:
     syserr("fork");
     break;
   }
  s->pid = childpid;
  if(waitforeditor) 
   {
     while((i = waitpid(childpid, (int*)NULL, 0)) < 0 && errno == EINTR);
     if(i < 0) syserr("waitpid");
     setsigs();		/* reset SIGINT and SIGCHLD handlers */
     finish_edit(s);
     free(s->descr);
     free(s->file);
     set_terminal();
     if (!edit_sess && *edit_end) {
         error = 0;
         parse_instruction(edit_end, 0, 0);
	 history_done = 0;
     }

     if (!s->builtin) {
         gotoxy(0, lines - 1); printf("\n%s", promptstr);
         redraw_input(); fflush(stdout);
         line0 = lines - 1;
     }
     free(s);

   }
  else
   {
     s->next = edit_sess;
     edit_sess = s;
   }
}

/*
 * split a command line into arguments:
 * return static NULL-terminated array of arguments
 * the line is assumed to be trashable.
 */
static char **split_line(line)
char *line;
{
  char *p;
  static char *args[MAXARGS + 1];
  int i;
  
  p = strtok(line, " ");
  for(i = 0; p != NULL && i < MAXARGS; p = strtok(NULL, " "),	i++)
     args[i] = p;
  args[i] = NULL;
  return args;
}

/*
 * Our child has snuffed it. check if it was an editor, and update the
 * session list if that is the case.
 */
void childdeathsig()
{
  int pid, ret;
  editsess **sp, *p;
  
  pid = wait(&ret);
  if(pid < 0) 
   {
     if(errno == ECHILD) return;	/* spurious signal? */
     syserr("wait");
   }
  if((ret & 0xff) != 0177) 
   {
     for(sp = &edit_sess; *sp && (*sp)->pid != pid; sp = &(*sp)->next);
     if(*sp) 
      {
	if((*sp)->key >= 0)	/* send back reply message? */
	   finish_edit(*sp);
	p = *sp; *sp = p->next;
	free(p->descr); free(p->file);
	free((char*)p);
      }
   } 
  else
     printf("process %d stopped (?)\n", pid);
  /*
   * system V signals are reset when caught. I hope it doesn't hurt to
   * reinstall it under BSD either.
   */
  signal(SIGCHLD, childdeathsig);

  /* send the edit_end message if this is the last editor... */ 
  if ((!edit_sess) && (*edit_end))
   {
     int osockfd = sockfd;  /* backup current socket fd */
     sockfd = main_sess();
     error = 0;
     parse_instruction(edit_end, 0, 0);  /* KLUDGE!!!  (oh well...) */
     history_done = 0;
     sockfd = osockfd;
   }
}

/*
 * send back edited text to server, or cancel the editing session if the
 * file was not changed.
 */
static void finish_edit(sp)
editsess *sp;
{
 int fd;
 char *msg, *text;
 int txtlen, hdrlen;
 struct stat sbuf;
 char keystr[16];
  
  if (sp->builtin) {
    if(unlink(sp->file) < 0) syserr("unlink edit file");
    return;
  }
  fd = open(sp->file, O_RDONLY);
  if(fd < 0) syserr("open edit file");
  if(fstat(fd, &sbuf) < 0) syserr("fstat");
  txtlen = sbuf.st_size;
  
  if(sbuf.st_mtime > sp->ctime || txtlen != sp->oldsize) 
   {
     /* file was changed by editor: send back result to server */
     
     text = (char*)malloc(txtlen + 1);	/* +1 is for possible LF */
     if(read(fd, text, txtlen) < txtlen)
     	syserr("read edit file");
     
     if(txtlen && text[txtlen - 1] != '\n') 
      {
	/* If the last line isn't LF-terminated, add an LF */
	text[txtlen] = '\n';
	txtlen++;
      }
     
     sprintf(keystr, "E%d\n", sp->key);
     msg = (char*)malloc(txtlen + 64);	/* 64 is safety margin */
     sprintf(msg, "%sE%d\n%s", MPI, (int)(txtlen + strlen(keystr)), keystr);
     hdrlen = strlen(msg);
     bcopy(text, msg + hdrlen, txtlen);
     free(text);

     msg[hdrlen + txtlen] = '\0';
     send_to_main_host(msg);
     PRINT_INT ("#completed session %s (%d)\n", sp->descr, sp->pid);
     
   } 
  else
   {
     /* file wasn't changed, cancel editing session */
     sprintf(keystr, "C%d\n", sp->key);
     msg = (char*)malloc(64);
     sprintf(msg, "%sE%d\n%s", MPI, (int) strlen(keystr), keystr);

     send_to_main_host(msg);
     PRINT_INT("#cancelled session %s (%d)\n", sp->descr, sp->pid);
   }
  
  free(msg);
  close(fd);
  if(unlink(sp->file) < 0) syserr("unlink edit file");
}
